#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2014 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl

from ._version import (__version__, __updated__)


__copyright__     = "Copyright (C) 2014 European Commission (JRC)"
__license__       = "EUPL 1.1+"
