#
# Copyright 2014 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
## Authoritative project's version-info
__version__ = "0.0.5"
__updated__ = "2014-11-12 16:22:55"
